/******************************************************************************
  HPFSRem.c
  Allows some removable media drives to retain removability with HPFS formatted
  disks. Tested on SyQuest's EZ135 Drive. Should work on all SyQuest drives.
  May work on other drives which can be formatted with HPFS.

  By: Jeffrey Jackowski
      jackowskij@email.uah.edu
      http://www.cs.uah.edu/cs/students/jjackows/hpfsrem
       (or just search for HPFSRem on Yahoo)

  Version 1     relased July 1996
  Version 1.1   relased August 1996
    Improves handling for drives without lock/unlock capability
    Better support for removable drives with drivers that neglect removability
    Error codes from OS/2 API now reported
    New disk status feature
    Revised documentation
  Version 1.11  relased October 1996
    Includes source code
    Now under the GNU General Public License, which follows:

 ************ License ************ 

    HPFSRem -- take out an HPFS disk without a system crash
    Copyright (C) 1996  Jeffrey Jackowski

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 ************ Note from the author (Jeffrey Jackowski)  ************ 

    I didn't bother to clean my code up -- its unchanged except for this
    comment and a couple other things. So, the rest of this program might
    not have the best comments and is doubtless not the best example of
    how to write code, but its not the worst, either.

    Anyway, it works on some drives and not others. Maybe someone out
    there can make it work better.

-------------------------------------------------------------------------------------------------------------------------------------------------

        Jeff Jackowski's version 1.11 had a few problems, most notably
        in the reporting of the status of the drive and media. I was suspicious
        that my SCSIZIP drive was always reported as not reporting it's lock
        status despite the WPS object for it having Lock/Unlack options.
        Looking at the code I saw why and fixed it. This version will correctly
        report the locked, unlocked, not reported or unlockable status of the drive.

        I've tried to get the remove option to eject the disk for you, but this
        doesn't work with my ZIP. Since the WPS eject option also doesn't work
        I have hopes that maybe this feature might work with some drives out
        there.

                                                        James Gibb, Nov 4, 1996

**************************************************************************/

#define VERSION 1.20

#define INCL_DOSDEVICES   /* Device values */
#define INCL_DOSERRORS    /* Error values */
#define INCL_DOSDEVIOCTL  /* IO control stuff */
#define INCL_DOSPROCESS
#define INCL_WINDIALOGS
#include <os2.h>
#include <stdio.h>

/* Drive status flags */
#define LOCK_STATUS               3
#define LOCK_STATUS_NOTLOCKABLE   0
#define LOCK_STATUS_LOCKED        1
#define LOCK_STATUS_NOTLOCKED     2
#define LOCK_STATUS_NOTREPORTED   3
#define MEDIA_STATUS              4
#define MEDIA_STATUS_PRESENT      4

/* These defines have been made by IBM, but are not always found in other
   compilers, like the GNU/EMX C compiler or WATCOM 10.6 */
#ifndef DSK_UNLOCKEJECTMEDIA
#define DSK_UNLOCKEJECTMEDIA 0x040
#endif
#ifndef DSK_GETLOCKSTATUS
#define DSK_GETLOCKSTATUS 0x066
#endif

// Some global variables - mainly used as dummys for parameters to Dos calls.
UCHAR zero = 0, IOCtlParam[2], IOCtlData;
ULONG len = 1;

void DispHelp() {  /* show a little help screen */
  printf("\nIncorrect syntax\n"
           "Please type \"HPFSRem <d> <f>\"\n"
           "where:\n"
           "\t<d> is the removable drive letter\n"
           "\t<f> is one of the following functions:\n"
           "\t\tR - remove disk\n"
           "\t\tI - insert disk\n"
           "\t\tRI - remove & insert disk\n"
           "\t\tS - drive info\n");
}

void DriveInfo(HFILE RemovableDisk) {
  int isRemovable, LockStatusValue, MediaPresentValue;
  APIRET ret;

  printf("\nDrive information as reported by OS/2:\n");

  //    Use DevIOCtl to see if the selected drive is reported as removable or not.
  DosDevIOCtl(RemovableDisk, IOCTL_DISK, DSK_BLOCKREMOVABLE, &zero, 2, &len, &IOCtlData, 1, &len);
  isRemovable = !IOCtlData;
  if (isRemovable) printf("\tRemovable\n");
  else printf("\tNon-removable  (but it may be removable anyway - this is reported from\n",
        "\t               the device driver )");

  //    Get the lock status of the drive.
  ret = DosDevIOCtl(RemovableDisk,IOCTL_DISK,DSK_GETLOCKSTATUS,&zero,2,&len,&IOCtlData,1,&len);
  if (ret == NO_ERROR)
   {
    LockStatusValue = IOCtlData & LOCK_STATUS;
    switch( LockStatusValue)
    {
    case LOCK_STATUS_NOTREPORTED:
       printf("\tLock status not reported by device driver.\n");
       break;
    case LOCK_STATUS_NOTLOCKABLE:
       printf("\tDrive is not lockable - beware of trying to use a file system that\n",
                "\t                        requires a controlled shutdown.\n");
       break;
    case LOCK_STATUS_LOCKED:
       printf("\tThe drive is locked.\n");
       break;
    case LOCK_STATUS_NOTLOCKED:
       printf("\tThe drive is not locked.\n");
       break;
    default:
       printf("Balls up somewhere.\n");
       break;
    }
    MediaPresentValue = IOCtlData & MEDIA_STATUS;
    switch(MediaPresentValue)
    {
    case MEDIA_STATUS_PRESENT:
       printf("\tMedia is in the drive.\n");
       break;
    default: 
       printf("\tNo media in the drive.\n");
       break;
    }
  }

  if (!isRemovable | ((ret == NO_ERROR) & (LockStatusValue != LOCK_STATUS_NOTLOCKABLE)))
    printf("\nDrive supports HPFS removability if it can be formatted with HPFS");
  else printf("\nDrive *may* not support HPFS removability");
  printf("\n  Consult documentation for details\n");
}

int Insert(HFILE hdlDrive)
 {
  APIRET ret;

  printf("Place a disk in the drive and press enter to continue.");
  getchar();       /* wait for key press */

  IOCtlParam[0] = 1;    /* lock command */
  ret = DosDevIOCtl(hdlDrive,IOCTL_DISK,DSK_UNLOCKEJECTMEDIA,&IOCtlParam,2,&len,&IOCtlData,1,&len);
  if (ret != NO_ERROR) return ret;

  //    Force the driver to shutdown and restart the file system for the new disk.
  ret = DosDevIOCtl(hdlDrive,IOCTL_DISK,DSK_REDETERMINEMEDIA,&zero,1,&len,&IOCtlData,1,&len);
  if (ret != NO_ERROR)
  {
    printf("Encountered an error when re-determining the new disk. Code is %i.",ret);
     return ret;
  }
  return NO_ERROR;
}

int Remove(HFILE hdlDrive)
  {
  APIRET ret;

  // The drive is already locked so this is the only process that can access it.
  // Flush the cache so no data is lost.
  ret = DosResetBuffer(hdlDrive);
  if (ret != NO_ERROR)               
        {
        printf("Encountered and error whilst flushing the cache. Code is %i.",ret);
        return -1;
        }

  // Force the file system (whatever it is - HPFS, FAT, TFS etc.) to close for this drive.
  ret = DosDevIOCtl(hdlDrive,IOCTL_DISK,DSK_REDETERMINEMEDIA,&zero,1,&len,&IOCtlData,1,&len);
  if (ret != NO_ERROR)
        {
        printf("Encountered and error whilst closing the file system. Code is %i.",ret);
        return -1;
        }

  // Unlock the drive (no need to check whether it needs unlocking, just get on and do it.
  IOCtlParam[0] = 0;    // Reserved
  IOCtlParam[1] = 0;    // Unlock the drive
  if (DosDevIOCtl(hdlDrive, IOCTL_DISK,DSK_UNLOCKEJECTMEDIA,&IOCtlParam,2,&len,&IOCtlData,1,&len) !=NO_ERROR)
     printf("Couldn't unlock the drive prior to ejecting it.\n");

  IOCtlParam[1] = 2;    // Eject the media
                        // This doesn't work for some hardware/driver combinations (eg. IOMega ZIP & Sequest drivers)
                        // but still doesn't report an error. Suck it and see.
  if (DosDevIOCtl(hdlDrive, IOCTL_DISK,DSK_UNLOCKEJECTMEDIA,&IOCtlParam,2,&len,&IOCtlData,1,&len) != NO_ERROR)
     printf("Couldn't eject the media from the drive.\n");

  printf("You may now remove the disk.\n");
  WinAlarm(HWND_DESKTOP, WA_NOTE);
  return NO_ERROR;
}

UnlockAndClose(HFILE hdlDrive)
{  // Unlock the drive and then close the handle
  if (DosDevIOCtl(hdlDrive,IOCTL_DISK,DSK_UNLOCKDRIVE,&zero,1,&len,&IOCtlData,1,&len) != NO_ERROR)
    printf("Error: could not unlock the drive");

  if (DosClose(hdlDrive) != NO_ERROR)
    printf("Error: could not close the drive");
}

int main(int numParams, char *params[])
{
  HFILE hdlDisk = -1L;
  ULONG action;
  UCHAR Drive[3], driveNumber;
  UCHAR RemoveIt = 0, InsertIt = 0, x;
  APIRET ret;

  printf("Removability for HPFS     Version %.2f\n"
         "  By: Jeffrey Jackowski   Copyright 1996\n"
         "  Modified by: James Gibb, November 1996\n", VERSION);

  if (numParams != 3) /* make certain correct number of parameters recived  */
                      /* 0 = prog. name, 1 = drive letter and 2 = function. */  
    {
    DispHelp();
    return -1;  /* give error code */
    }

  Drive[0] = params[1][0];
  if (Drive[0] > 'Z') driveNumber = Drive[0] - 'a'; /* Convert drive letter to */
  else driveNumber = Drive[0] - 'A';                /* drive number            */
  if (driveNumber > 26)          /* check for invalid input, i.e. not a letter */
    {
    DispHelp();
    return -1;
    }

  //    Now we have a consistent drive letter, make a "d:\0" string
  Drive[0] = 'a' + driveNumber;
  Drive[1] = ':';
  Drive[2] = 0;

  //    Get a device handle for the specified disk
  ret = DosOpen(Drive, &hdlDisk, &action, 0L, 0, OPEN_ACTION_OPEN_IF_EXISTS, OPEN_FLAGS_DASD |
        OPEN_FLAGS_FAIL_ON_ERROR | OPEN_SHARE_DENYREADWRITE | OPEN_ACCESS_READWRITE, 0);
  switch (ret) {
  case ERROR_SHARING_VIOLATION:
     printf("Encountered a sharing violation whilst trying to get the disk.\n");
     break;
  case NO_ERROR:
     break;
  default: 
     printf("Failed to open the disk, Returned error was %i.",ret);
     return -1;
    break;
  } /* endswitch */

  for (x = 0; x < 2; x++)    /* figure out user's instructions (2nd param) */
    {                        /* may be R, I or RI or S                     */
    switch (params[2][x])
      {
      case 'R':              /* remove disk */
      case 'r':
        RemoveIt = 1;
        break;
      case 'I':              /* insert disk */
      case 'i':
        InsertIt = 1;
        break;
      case 'S':              /* display info about disk */
      case 's':
        DriveInfo(hdlDisk);
        return 0;
      case 0:        /* takes care of null termination if */
        break;       /*  parameter is 1 char long         */
      default:
        DispHelp();  /* invalid option - help user out */
        return -1;
      }   // End switch
    }     // end for

  // RemovableDisk now contains the handle of the drive so we can use that instead of the drive number in the IOCtl call.
  // Attempt to lock the drive so only this process can use it.
  if (DosDevIOCtl(hdlDisk,IOCTL_DISK,DSK_LOCKDRIVE,&zero,1,&len,&IOCtlData,1,&len) != NO_ERROR)
    {
    printf("Error: could not lock the drive. Something on it may be open");
    return -1;
    }

  if (RemoveIt)                              /* remove disk */
    {
    ret = Remove(hdlDisk);
    if (ret != NO_ERROR)
      {
      printf("Error: could not prepare disk for removal, code %i\n",ret);
      UnlockAndClose(hdlDisk);
      return -1;
      }
    }

  if (InsertIt)                               /* insert disk */
    {
    ret = Insert(hdlDisk);
    if (ret != NO_ERROR)
      {
      printf("Error: cannot allow disk to be inserted, code %i\n",ret);
      UnlockAndClose(hdlDisk);
      return -1;
      }
    }

  UnlockAndClose(hdlDisk);
                                   
  printf("Task completed successfully\n");
  return 0;
}
